// RootflyGraphics.cpp: implementation of the RootflyGraphics class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "rootfly.h"
#include "RootflyGraphics.h"
#include "DataStructures.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


COLORREF ROOTFLY_RGB_RED =         RGB(255,0,0);
COLORREF ROOTFLY_RGB_GREEN =       RGB(0,255,0);
COLORREF ROOTFLY_RGB_BLUE =        RGB(0,0,255);
COLORREF ROOTFLY_RGB_BLACK =       RGB(0,0,0);
COLORREF ROOTFLY_RGB_DARK_BROWN =  RGB(139, 69, 19 );
COLORREF ROOTFLY_RGB_LIGHT_BROWN = RGB(205, 133, 63);
COLORREF ROOTFLY_RGB_WHITE =       RGB(255, 255, 255);
COLORREF ROOTFLY_RGB_SYS_GRAY =    ::GetSysColor(COLOR_3DFACE);

// Note:  Roots have several drawing modes:
//
//                            width     color    draw points?   ID outline?
//   Normal (live)              1         NL         no             no
//   Normal (dead)              1         ND         no             no
//   Mouse over (live)          1         SL         no             no
//   Mouse over (dead)          1         SD         no             no
//   Selected   (live)          2         SL         yes            yes
//   Selected   (dead)          2         SD         yes            yes
//   New root being drawn       1         SL         no             no
//

const int RootflyGraphics::m_point_size = 6;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
RootflyGraphics::RootflyGraphics()
{

}

RootflyGraphics::~RootflyGraphics()
{

}

void RootflyGraphics::Initialize( const ImageClientConverter &icc, const Settings &s, const RootflyData &rfd )
{
  rfy_dlg_conv = &icc;
  rfy_dlg_settings = &s;
  rfy_dlg_data = &rfd;
}

COLORREF RootflyGraphics::GetRootColor(const RootInstance& inst, bool selected) const
{
  if (selected)
  {
	  if (inst.IsAlive())  return rfy_dlg_settings->GetRootDrawingColorLiveSelected();
	  else                 return rfy_dlg_settings->GetRootDrawingColorDeadSelected();
  }
  else
  {
    if (inst.IsAlive())  return rfy_dlg_settings->GetRootDrawingColorLive();
  	else                 return rfy_dlg_settings->GetRootDrawingColorDead();
  }  
}

int RootflyGraphics::GetRootLineDrawingWidth( bool selected ) const
{
  return selected ? 2 : 1;
}

// draws the diameter
// 'selected':  whether to draw root with 'selected' color
void RootflyGraphics::DrawDiameter(const RootInstance& inst, COLORREF color, bool draw_point, CDC& dc) const
{
  const RDiameter *diam = inst.GetDiameter(0);
  RPoint dpoint = diam->r_point;
  CPoint cpt1,cpt2;
  double rad = diam->radius;
  cpt1 = rfy_dlg_conv->ImageToClient( CPoint(dpoint.x - rad, dpoint.y - rad), dpoint.window );
  cpt2 = rfy_dlg_conv->ImageToClient( CPoint(dpoint.x + rad, dpoint.y + rad), dpoint.window );
  CRect rect(cpt1.x, cpt1.y, cpt2.x, cpt2.y);
  
  CPen pen(PS_SOLID, 1, color);
  CPen* oldpen = dc.SelectObject(&pen);
  CBrush brush;
  brush.CreateSolidBrush( RootColorToRgb( inst.GetColor() ) );
  dc.SelectObject(brush);   
  dc.Ellipse( rect );
  dc.SelectObject(oldpen);  

  if (draw_point) 
  { // draw extra point for resizing
    const RPoint point = RPoint(dpoint.x, dpoint.y - rad, dpoint.window);
    DrawPoint(point, dc);
  }
}

void RootflyGraphics::DrawRootId(const Root& root, 
                                 bool selected, 
                                 int current_window, 
                                 int session, 
                                 CDC& dc) const
{
  // displays the root id, checks that the current window is the same as the primary window for the RootInstance, if it is not
  // then it will show the RootInstance as "WindowID(PrimaryWindow
  int window = root.GetPrimaryWindow();
  CString fmt = "%d";
  if (window != current_window) fmt += "*";
  CString str;
  str.Format(fmt, root.GetDisplayableID());

  const RootInstance *inst = root.GetRootInstance( session );
  if (!inst)  return;
  CRect main = rfy_dlg_conv->GetClientRect( window );
  const std::vector<RPoint>* line = inst->GetAllPoints();
  if (!line)  return;
  if (line->size() > 0)
  {
    CPoint conv_point = rfy_dlg_conv->ImageToClient( line->at(0) );
    CPoint topleft(conv_point.x-15, conv_point.y-15);
    if (topleft.x < main.left) topleft.x = main.left;
    if (topleft.y < main.top ) topleft.y = main.top;
    dc.SetTextColor(ROOTFLY_RGB_BLACK);
    dc.TextOut(topleft.x, topleft.y, str);
    if (selected)
    { // draw outline around root id
      CSize sz = dc.GetTextExtent(str);
      COLORREF color = GetRootColor(*inst, selected);
      CBrush brush( color );
      CRect rect(topleft.x, topleft.y, topleft.x + sz.cx, topleft.y + sz.cy);
      rect.InflateRect(1, 1, 1, 1);
      dc.FrameRect( rect, &brush );
    }
  }
}

// RootColorToRgb function
COLORREF RootflyGraphics::RootColorToRgb( RootInstance::Color tempcolor ) const
{
  switch (tempcolor)
  {
  case RootInstance::COLOR_WHITE:       return ROOTFLY_RGB_WHITE;
  case RootInstance::COLOR_LIGHTBROWN:  return ROOTFLY_RGB_LIGHT_BROWN;
  case RootInstance::COLOR_DARKBROWN:   return ROOTFLY_RGB_DARK_BROWN;
  case RootInstance::COLOR_BLACK:       return ROOTFLY_RGB_BLACK;
  default:
    assert(0);  // root should have a color
    return ROOTFLY_RGB_WHITE;
  }
}

// draws points on the selected root and diameter
void RootflyGraphics::DrawPoint(const RPoint& point, CDC& dc) const
{
  int size = m_point_size / 2;    // halfsize
  
  CPoint cpt = rfy_dlg_conv->ImageToClient( point );
  CRect rect(cpt.x - size, cpt.y - size, cpt.x + size, cpt.y + size);
  CPen pen(PS_SOLID, 1, ROOTFLY_RGB_BLACK);
  CPen* oldpen = dc.SelectObject(&pen);
  CBrush brush;
  brush.CreateSolidBrush(ROOTFLY_RGB_WHITE);
  dc.SelectObject(brush);
  dc.Rectangle(rect);
  dc.SelectObject(oldpen);   
}

// draws points on the selected root
void RootflyGraphics::DrawPoints(const std::vector<RPoint>& line, CDC& dc) const
{
  int size = m_point_size / 2;    // halfsize
  
  CPen pen(PS_SOLID, 1, ROOTFLY_RGB_BLACK);
  CPen* oldpen = dc.SelectObject(&pen);
  CBrush brush;
  brush.CreateSolidBrush(ROOTFLY_RGB_WHITE);
  CBrush* oldbrush = dc.SelectObject(&brush);
  std::vector<RPoint>::const_iterator pt;
  for (pt = line.begin() ; pt != line.end() ; pt++)
  {
    CPoint cpt = rfy_dlg_conv->ImageToClient( *pt );
    CRect rect(cpt.x - size, cpt.y - size, cpt.x + size, cpt.y + size);
    dc.Rectangle(rect);
  }
  dc.SelectObject(oldpen);   
  dc.SelectObject(oldbrush);   
}

void RootflyGraphics::DrawPolyline( const std::vector<RPoint>& line, COLORREF c, CDC& dc ) const
{
  DrawPolyline(line, c, 1, dc);
}

void RootflyGraphics::DrawPolyline( const std::vector<RPoint>& line, COLORREF c, int width, CDC& dc ) const
{	 
  if ( line.size() >= 2 )
  {
    CPen pen(PS_SOLID, width, c );
    CPen* oldpen = dc.SelectObject(&pen);
    vector<RPoint>::const_iterator p = line.begin();
    CPoint pt;
    
    pt = rfy_dlg_conv->ImageToClient( *p );
    dc.MoveTo(pt.x, pt.y);
    for ( ; p != line.end() ; p++)
    {
      pt = rfy_dlg_conv->ImageToClient( *p );
      dc.LineTo(pt.x, pt.y);
    } 
    
    dc.SelectObject(oldpen);
  }  
}

void RootflyGraphics::ExtendPolyline( const RootInstance& ri, 
                                      bool extend_from_front,
                                      COLORREF color,
                                      int width,
                                      CPoint mouse_point,
                                      CDC& dc) const
{
  const std::vector<RPoint>* line = ri.GetAllPoints();
  if (line && line->size() > 0)
  {
    std::vector<RPoint> xtend;
    xtend.push_back( extend_from_front ? (*line)[0] : line->back() );
    xtend.push_back( rfy_dlg_conv->ClientToImage( mouse_point ) );
    DrawPolyline(xtend, color, width, dc);
  }
}

// Draw a single root
// (Code should eventually be refactored to use this function.
//  First, though, we may need to make slight modifications to the function to ensure
//  that its functionality is as desired.)  -- STB 4/21/08
void RootflyGraphics::DrawSingleRoot(CDC& dc, 
                                     const Root &root, 
                                     int current_window,
                                     int session,
                                     bool new_root_being_drawn,
                                     bool selected,
                                     bool display_diameter,
                                     bool draw_resizing_point,  // for diameter
                                     bool display_root_id) const
{
  const RootInstance* ri = root.GetRootInstance( session );
  if (ri)
  {
    COLORREF color = GetRootColor(*ri, selected);
    int width = GetRootLineDrawingWidth(selected);
    DrawPolyline( *ri->GetAllPoints(), color, width, dc );
    DrawPoints( *ri->GetAllPoints(), dc);
    if (display_diameter) DrawDiameter( *ri, color, draw_resizing_point, dc );
    if (display_root_id)  DrawRootId(root, selected, current_window, session, dc);
//    ExtendPolyline( *ri, extend_from_front, color, mouse_point, dc);
  }
}

// Draws the selected root
void RootflyGraphics::DrawSelectedRoot(CDC& dc, int selected_id, int mode, const CPoint& mouse_point, bool xtend_tail) const
{
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  const Experiment& exp = rfy_dlg_data->GetExperiment();
  const Tube *tube = exp.GetTubeByName(rfy_dlg_data->GetApplicationData().GetCurrentTubeName());
  if (!tube)  return;
  const Root *root = tube->GetRoot( selected_id );
  if (!root)  return;
  const RootInstance *inst = root->GetRootInstance(rfy_dlg_data->GetApplicationData().GetCurrentSession());
  if (!inst)  return;
  const std::vector<RPoint> *line = inst->GetAllPoints();
  if (!line)  return;
  const RDiameter *diam = inst->GetDiameter(0);
  if (!diam)  return;
 
  bool selected = true;
  COLORREF color = GetRootColor(*inst, selected);
  DrawPolyline(*line, color, 2, dc);
  if (rfy_dlg_settings->RootIdIsEnabled())
  {
    DrawRootId(*root, selected, rfy_dlg_data->GetApplicationData().GetCurrentWindow(), rfy_dlg_data->GetApplicationData().GetCurrentSession(), dc);
  }
  if (rfy_dlg_settings->ShowDiametersIsEnabled())
  {    
    DrawDiameter(*inst, color, (mode == MODE_ROOTSELECTED), dc);
  }
  if ( mode == MODE_ROOTSELECTED )
  {
    DrawPoints( *line, dc );
  }  
  if ( mode == MODE_EXTENDINGENDPOINT )
  {
    int width = GetRootLineDrawingWidth( selected );
    ExtendPolyline( *inst, !xtend_tail, color, width, mouse_point, dc);
  }
}

// Draws all the roots in the data structure
void RootflyGraphics::DrawAllRoots(CDC& dc, int mode, bool zoomed) const
{
  if ( mode == MODE_NOFILELOADED ) return;  
  if ( mode == MODE_NOIMAGEDISPLAYED ) return;  
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  const Experiment& exp = rfy_dlg_data->GetExperiment();   
  const Tube *tube = exp.GetTubeByName(rfy_dlg_data->GetApplicationData().GetCurrentTubeName());
  if (tube == NULL) return;
  const vector<Root> *root = tube->GetAllRoots();
  if (root == NULL) return;
  
  
  // Draws roots
  for ( int k = 0; k <= exp.GetMaxWindows(); k++ )
  {
    int window = k;  // draw all windows regardless of position
//      int window = rfy_dlg_data->GetApplicationData().GetCurrentWindow() + k - 1;
    if ( window < 1  ) continue;
    for ( int i = 0; i < root->size(); i++ )
    {
      if (root->at(i).GetPrimaryWindow() == window)
      {
        const RootInstance *inst = root->at(i).GetRootInstance(rfy_dlg_data->GetApplicationData().GetCurrentSession());
        if (inst == NULL)  continue;           
        if (inst->IsAlive() || rfy_dlg_settings->DeadRootsIsEnabled() )
        {
          bool highlight = ( mode == MODE_STARTINGMOVEALL || mode == MODE_MOVINGALLROOTS ) 
                              && window == rfy_dlg_data->GetApplicationData().GetCurrentWindow();
          DrawPolyline(*inst->GetAllPoints(), GetRootColor(*inst, highlight), dc);
          if (rfy_dlg_settings->RootIdIsEnabled())
          {
            bool selected = false;
            DrawRootId(root->at(i), selected, rfy_dlg_data->GetApplicationData().GetCurrentWindow(), rfy_dlg_data->GetApplicationData().GetCurrentSession(), dc);
          }
          if ( rfy_dlg_settings->ShowDiametersIsEnabled())
          { 
           //DrawCircle(bounds, inst->IsAlive()?m_set_data->GetRootDrawingColor():m_set_data->GetRootDrawingColorDead(), inst->GetColor());
            bool highlight = ( mode == MODE_MOVINGALLROOTS || mode == MODE_STARTINGMOVEALL ) && window == rfy_dlg_data->GetApplicationData().GetCurrentWindow();
            DrawDiameter(*inst, GetRootColor(*inst, highlight), false, dc);
          } 
        }
      }
    }
  } 
}

// Draws the new root that is currently being created
void RootflyGraphics::DrawCurrentRoot(CDC& dc, int mode, const RootInstance &ri, const CPoint& mouse_point) const
{
  const std::vector<RPoint>* line = ri.GetAllPoints();
  COLORREF color = GetRootColor(ri, true);
  int width = 1;

  if ( mode == MODE_DRAWINGNEWROOT )
  {	
    DrawPolyline(*line, color, width, dc);    
    ExtendPolyline(ri, false, color, width, mouse_point, dc);
  }  
  
  //Draw the diameter that is being created
  if ( mode == MODE_SIZINGDIAMETER || mode == MODE_LOCATINGDIAMETER  )
  {
    DrawPolyline(*line, color, width, dc);
    DrawDiameter(ri, GetRootColor(ri, true), false, dc);    
  }   
}

// Sets the clip region of the DC to either 
//    - the main image (in regular mode), or
//    - the union of the three images (in zoomed out mode).
// Should only need to call this function once before starting all the drawing.
// Note:  SelectClipRgn only seems to work for the real CDC, not for the MemDC.
void RootflyGraphics::SetClipRegion(CDC& dc, bool zoomed)
{
  CRgn clip_rgn;
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  if ( zoomed )
  {
    CRect rect, rect1, rect2, rect3;
    int w = rfy_dlg_data->GetApplicationData().GetCurrentWindow();
    rect1 = rfy_dlg_conv->GetClientRect( w-1 );
    rect2 = rfy_dlg_conv->GetClientRect( w   );
    rect3 = rfy_dlg_conv->GetClientRect( w+1 );
    
    rect.UnionRect(rect1, rect2);
    rect.UnionRect(rect, rect3);
    clip_rgn.CreateRectRgnIndirect(&rect);
  }
  else
  {
    CRect rect;
    int w = rfy_dlg_data->GetApplicationData().GetCurrentWindow();
    rect = rfy_dlg_conv->GetClientRect( w );
    clip_rgn.CreateRectRgnIndirect(&rect);
  }
  //  clip region
  dc.SelectClipRgn(&clip_rgn); 
}

void RootflyGraphics::SetClipRegion(CDC& dc, const CRect& rect)
{
  CRgn clip_rgn;
  clip_rgn.CreateRectRgnIndirect(&rect);
  dc.SelectClipRgn(&clip_rgn); 
}

void RootflyGraphics::FillImageDisplayAreaWithGray(CDC& dc, const CRect &rect)
{
  // fill image area with gray
  CBrush background(ROOTFLY_RGB_SYS_GRAY);
  dc.FillRect(rect, &background);
}

bool RootflyGraphics::RootIsInCurrentWindow( const RootInstance &inst ) const
{
  const vector<RPoint> *points = inst.GetAllPoints();
  const ApplicationData& app_data = rfy_dlg_data->GetApplicationData();
  for ( int i = 0; i < points->size() -1 ; i++ )
  {
    if ( points->at(i).window == rfy_dlg_data->GetApplicationData().GetCurrentWindow() )
      return true;
  }
  return false;
}